#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define OP_FNCT 0x0UL

#define FNCT_ADD 0x0UL
#define FNCT_SUB 0x1UL
#define FNCT_SLL 0x2UL
#define FNCT_SRR 0x3UL
#define FNCT_AND 0x4UL
#define FNCT_OR  0x5UL
#define FNCT_XOR 0x6UL
#define FNCT_NOT 0x7UL
#define FNCT_FADD  0x8UL
#define FNCT_FSUB  0x9UL
#define FNCT_FMUL  0xAUL
#define FNCT_FINV  0xCUL
#define FNCT_FSQRT 0xDUL
#define FNCT_RD 0xEUL
#define FNCT_WR 0xFUL

#define FNCT_SIN 0x10UL
#define FNCT_COS 0x11UL
#define FNCT_ATAN 0x12UL
#define FNCT_BR 0x13UL

#define OP_ADDI 0x1UL

#define OP_LOAD  0x2UL
#define OP_STORE 0x3UL

#define OP_BL   0xAUL
#define OP_BE   0x8UL
#define OP_BLE  0xBUL
#define OP_BNE  0x9UL
#define OP_FBL   0xCUL
#define OP_FBLE  0xDUL

#define OP_JR   0xEUL
#define OP_JAL  0xFUL

#define REG_SIZE 32
#define MEM_SIZE (1 << 20)
unsigned long reg[REG_SIZE];
unsigned long mem[MEM_SIZE];
char *status = NULL;

unsigned long op_count[FNCT_BR + OP_JAL + 1];

#define LABEL_MAX 1024
unsigned long	asm_label_pos [LABEL_MAX];
char		asm_label_name[LABEL_MAX][128];
unsigned long	label_count[LABEL_MAX];
char		label_name [LABEL_MAX][128];

unsigned long br_skip = 0;

void print_report(){
  int i;
  
  printf("Operations executed\n");
  printf("----------------------------------------------------------------\n");
  printf("ADD \t : %ld\n", op_count[FNCT_ADD]);
  printf("SUB \t : %ld\n", op_count[FNCT_SUB]);
  printf("SLL \t : %ld\n", op_count[FNCT_SLL]);
  printf("SRR \t : %ld\n", op_count[FNCT_SRR]);
  printf("AND \t : %ld\n", op_count[FNCT_AND]);
  printf("OR  \t : %ld\n", op_count[FNCT_OR ]);
  printf("XOR \t : %ld\n", op_count[FNCT_XOR]);
  printf("NOT \t : %ld\n", op_count[FNCT_NOT]);
  printf("FADD  \t : %ld\n", op_count[FNCT_FADD ]);
  printf("FSUB  \t : %ld\n", op_count[FNCT_FSUB ]);
  printf("FMUL  \t : %ld\n", op_count[FNCT_FMUL ]);
  printf("FINV  \t : %ld\n", op_count[FNCT_FINV ]);
  printf("FSQRT \t : %ld\n", op_count[FNCT_FSQRT]);
  printf("RD \t : %ld\n", op_count[FNCT_RD]);
  printf("WR \t : %ld\n", op_count[FNCT_WR]);
  printf("SIN  \t : %ld\n", op_count[FNCT_SIN ]);
  printf("COS  \t : %ld\n", op_count[FNCT_COS ]);
  printf("ATAN \t : %ld\n", op_count[FNCT_ATAN]);
  printf("BR \t : %ld\n", op_count[FNCT_BR]);
  
  printf("\n");
  
  printf("Label called from JAL\n");
  printf("----------------------------------------------------------------\n");
  for (i = 0; label_count[i] > 0; i++)
    printf("%s \t: %ld\n", label_name[i], label_count[i]);
  
}

void init_report(char *p){
  int i;
  FILE *fp_asm;
  
  for (i = 0; i < FNCT_BR + OP_JAL + 1; i++) op_count[i] = 0;
  
  sprintf(p, "-a");
  fp_asm = popen(p - 256, "r");
  
  for(i = 0; !feof(fp_asm); i++){
    fscanf(fp_asm, "%ld", &asm_label_pos [i]);
    fscanf(fp_asm, "%s" , &asm_label_name[i]);
  }
  
  for(i = 0; i < LABEL_MAX; i++) label_count[i] = 0;  
}

void count_label(char *s) {
  int i;
  
  for (i = 0; label_count[i] > 0; i++)
    if (strcmp(label_name[i], s) == 0) {
      label_count[i]++;
      return;
    }
  
  strcpy(label_name[i], s);
  label_count[i] = 1;
}

static unsigned long S(unsigned long l) {
  return l >> 31;
}
static unsigned long E(unsigned long l) {
  return l >> 23 & 0xFF;
}
static unsigned long F(unsigned long l) {
  return l & 0x7FFFFF | (l >> 23 & 0xFFL ? 0x800000L : 0L);
}
static unsigned long L(unsigned long s, unsigned long e, unsigned long f) {
  int i;
  for (i = 0; i < 255; i++) {
    if (f & 0x1000000UL) break;
    e -= 1;
    f <<= 1;    
  }
  e = e < 0xFFUL ? e + 1UL & 0xFFUL : 0;
  f = f >> 1 & 0x7FFFFFUL;
  return s << 31 | e << 23 | f;
}

unsigned long calc(unsigned long pc, FILE *fp_read, FILE *fp_write) {  

  const unsigned long op = mem[pc] >> 26;
  const unsigned long rs = mem[pc] >> 21 & 0x1FUL;
  const unsigned long rt = mem[pc] >> 16 & 0x1FUL;
  const unsigned long rd = mem[pc] >> 11 & 0x1FUL;
  const unsigned long shft = mem[pc] >> 6 & 0x1FUL;
  const unsigned long fnct = mem[pc] & 0x3FUL;
  const unsigned long oft = mem[pc] & 0xFFFFUL | (mem[pc] & 0x8000UL ?  ~0xFFFFUL : 0UL);
  const unsigned long tgt = mem[pc] & 0x1FFFFFL | (mem[pc] & 0x100000UL ?  ~0x1FFFFFUL : 0UL);

  op ? op_count[FNCT_BR + op]++ : op_count[fnct]++;
  
  switch (op) {
     
    /**********************************/
    /* GENERAL OPERATIONS             */
    /**********************************/
  case OP_FNCT:
    switch(fnct) {
    case FNCT_ADD:
      reg[rd] = reg[rs] + reg[rt] & 0xFFFFFFFF;
      return 1UL;
    case FNCT_SUB:
      reg[rd] = reg[rs] - reg[rt] & 0xFFFFFFFF;
      return 1UL;
    case FNCT_SLL:
      reg[rd] = (signed)reg[rs] << shft & 0xFFFFFFFF;
      return 1UL;
    case FNCT_SRR:
      reg[rd] = (signed)reg[rs] >> shft & 0xFFFFFFFF;
      return 1UL;
    case FNCT_AND:
      reg[rd] = reg[rs] & reg[rt];
      return 1UL;
    case FNCT_OR:
      reg[rd] = reg[rs] | reg[rt];
      return 1UL;
    case FNCT_XOR:
      reg[rd] = reg[rs] ^ reg[rt];
      return 1UL;
    case FNCT_NOT:
      reg[rd] = ~reg[rs] & 0xFFFFFFFF ^ reg[rt];
      return 1UL;
    case FNCT_FADD:
      if (E(reg[rs]) > E(reg[rt])
	  || E(reg[rs]) == E(reg[rt]) && F(reg[rs]) >= F(reg[rt])) {
	reg[rd]
	  = L(S(reg[rs]),
	      E(reg[rs]),
	      E(reg[rs]) - E(reg[rt]) > 23 ?
	      F(reg[rs]) :
	      S(reg[rs]) ^ S(reg[rt]) ?
	      F(reg[rs]) - (F(reg[rt]) >> E(reg[rs]) - E(reg[rt])) :
	      F(reg[rs]) + (F(reg[rt]) >> E(reg[rs]) - E(reg[rt])));
      } else {
	reg[rd]
	  = L(S(reg[rt]),
	      E(reg[rt]), 
	      E(reg[rt]) - E(reg[rs]) > 23 ?
	      F(reg[rt]) :
	      S(reg[rs]) ^ S(reg[rt]) ?
	      F(reg[rt]) - (F(reg[rs]) >> E(reg[rt]) - E(reg[rs])) :
	      F(reg[rt]) + (F(reg[rs]) >> E(reg[rt]) - E(reg[rs])));
      }
      return 1UL;
    case FNCT_FSUB:
      if (E(reg[rs]) > E(reg[rt])
	  || E(reg[rs]) == E(reg[rt]) && F(reg[rs]) >= F(reg[rt])) {
	reg[rd]
	  = L(S(reg[rs]),
	      E(reg[rs]),
	      E(reg[rs]) - E(reg[rt]) > 23 ?
	      F(reg[rs]) :
	      S(reg[rs]) ^ S(reg[rt]) ?
	      F(reg[rs]) + (F(reg[rt]) >> E(reg[rs]) - E(reg[rt])) :
	      F(reg[rs]) - (F(reg[rt]) >> E(reg[rs]) - E(reg[rt])));
      } else {
	reg[rd]
	  = L(!S(reg[rt]),
	      E(reg[rt]),
	      E(reg[rt]) - E(reg[rs]) > 23 ?
	      F(reg[rt]) :
	      S(reg[rs]) ^ S(reg[rt]) ?
	      F(reg[rt]) + (F(reg[rs]) >> E(reg[rt]) - E(reg[rs])) :
	      F(reg[rt]) - (F(reg[rs]) >> E(reg[rt]) - E(reg[rs])));
      }
      return 1UL;
    case FNCT_FMUL:
      reg[rd]
	= L(S(reg[rs]) ^ S(reg[rt]),
	    E(reg[rs]) + E(reg[rt]) - 128,
	    (((long long)F(reg[rt]) * (F(reg[rs]) % 0x1000UL)) >> 23)
	    + (((long long)F(reg[rt]) * (F(reg[rs]) / 0x1000UL)) >> 11));
      return 1UL;
    case FNCT_FINV:
      {
	const unsigned long addr
	  = F(reg[rs]) >> 13 & 0x3FEUL | 0x401UL;
	const unsigned long long tbl
	  = 0x400000ULL / addr - 0x200000ULL / addr & 0x3FFUL | 0x400UL;
	reg[rd]
	  = L(S(reg[rs]),
	      255 - E(reg[rs]),
	      tbl * (0x1000006UL - F(reg[rs]) / 2 * tbl / 0x400UL) / 0x400UL);	    
      }
      return 1UL;
    case FNCT_FSQRT:
      if (F(reg[rs]) & 0x800000) {
	const unsigned long addr
	  = (F(reg[rs]) >> 14 & 0x1FEUL | 0x201UL) << (E(reg[rs]) & 0x1UL);
	const unsigned long long itbl
	  = (int)(1. / sqrt((double)addr / 0x200UL) * 0x1000UL + 0.5) >> 1 & 0x3FF | 0x400;
	const unsigned long long tbl
	  = 0x100000000ULL / itbl - 0x80000000ULL / itbl - 1 & 0x0FFFFFUL | 0x100000;
	reg[rd]
	  = L(0,
	      E(reg[rs]) / 2 + 64,
	      tbl * 0x4UL
	      + (F(reg[rs]) / 2 * itbl / 0x800 << E(reg[rs]) % 2));
      } else {
	reg[rd] = 0L;
      }
      return 1UL;
    case FNCT_RD:
      return (reg[rd] = fgetc(fp_read)) != (unsigned)EOF;
    case FNCT_WR:
      fputc(reg[rs], fp_write);
      return 1;
    case FNCT_SIN:
      {
	float f = *(float *)&reg[rs];
	f = sin(f / 2.) * 2.;
	reg[rd] = *(long *)&f;
	return 1UL;
      }
    case FNCT_COS:
      {
	float f = *(float *)&reg[rs];
	f = cos(f / 2.) * 2.;
	reg[rd] = *(long *)&f;
	return 1UL;
      }
    case FNCT_ATAN:
      {
	float f = *(float *)&reg[rs];
	f = atan(f / 2.) * 2.;
	reg[rd] = *(long *)&f;
	return 1UL;
      }
    case FNCT_BR:
      if(br_skip == 0)
        status = "";
      else
        br_skip--;
      return 1UL;
    default:
      status = "Message: Undefined function detected.\n";
      return 1L;
    }
    
  case OP_ADDI:
    reg[rt] = reg[rs] + oft & 0xFFFFFFFF;
    return 1UL;
    
    /**********************************/
    /* MEMORY OPERATIONS              */
    /**********************************/
  case OP_LOAD:
    {
      const unsigned long addr = (reg[rs] + oft) % MEM_SIZE;
      reg[rt] = mem[addr];
      if (addr < 4
	  || addr >= 0x40000 && addr < reg[1] % MEM_SIZE - 4
	  || addr >= reg[2] % MEM_SIZE) {
	status = "Message: Exception detected.\n";
      }
      return 1UL;
    }
  case OP_STORE:
    {
      const unsigned long addr = (reg[rs] + oft) % MEM_SIZE;
      if (addr < 4
	  || addr >= 0x40000 && addr < reg[1] % MEM_SIZE - 4
	  || addr >= reg[2] % MEM_SIZE) {
	status =  "Message: Exception detected.\n";
      }
      mem[addr] = reg[rt];
#ifdef DUMP
      printf("%ld %08X\r\n", addr, reg[rt]);
#endif
      return 1UL;
    }
    
    /**********************************/
    /* BRANCH OPERATIONS              */
    /**********************************/
  case OP_BL  :
    return (signed)reg[rs] <  (signed)reg[rt] ? oft : 1;
  case OP_BE  :                  
    return (signed)reg[rs] == (signed)reg[rt] ? oft : 1;
  case OP_BLE :                  
    return (signed)reg[rs] <= (signed)reg[rt] ? oft : 1;
  case OP_BNE :                  
    return (signed)reg[rs] != (signed)reg[rt] ? oft : 1;
  case OP_FBL  :
    return ((signed)reg[rs] < (signed)reg[rt]) ^ S(reg[rs]) & S(reg[rt]) ? oft : 1UL;
  case OP_FBLE :                 
    return ((signed)reg[rs] <= (signed)reg[rt]) ^ S(reg[rs]) & S(reg[rt])  ? oft : 1UL;

    /**********************************/
    /* JUMP OPERATIONS                */
    /**********************************/
  case OP_JR:
    return reg[rs] + tgt - pc;
  case OP_JAL:
    {
      int i;
      
      for (i = 0; 1; i++)
        if (asm_label_pos[i] == pc) {
          count_label(asm_label_name[i]);
          break;
        }
      
    }
    reg[31] = pc;		//1 or 4?
    return tgt;      
    
  default:
    status = "Message: Undefined operation detected\n";
    return 1UL;
    
  }   
  
}

int main(int argc, char *argv[])
{
  int i;
  unsigned long pc, dpc;  
  FILE *fp_execute, *fp_read, *fp_write;
  char *p;

  if (argc != 4
      || (fp_execute = fopen(argv[1], "r")) == NULL
      || (fp_read = fopen(argv[2], "r")) == NULL
      || (fp_write = fopen(argv[3], "w")) == NULL) {
    printf("Usage: simulate execute_file read_file write_file\n");
    return 1;
  }

  p = (char *)mem + fread(mem, 1, sizeof(long) * MEM_SIZE, fp_execute);
  
  init_report(p);
  
  pc = 0;
  dpc = 0;
  do {
    /* ߂sāAPC̑擾B */
    dpc = calc(pc, fp_read, fp_write);
    /* PCvZ */
    pc = (pc + dpc) % MEM_SIZE;
    if (status != NULL) {
      int i, j, len;
      printf("----------------------------------------------------------------\n");
      /* 󋵂\ */
      printf("%s", status);
      status = "";
      printf("----------------------------------------------------------------\n");
      /* I[256oCgɃAZũR}hC}Ă */
      /* R}hCIvVt(Ȃ) */
      sprintf(p, "-d -b %ld", pc);
      /* AZuɕt߂̗lq\ */
      system(p - 256);
      printf("----------------------------------------------------------------\n");
      /* R}h󂯕t */
      while(1) {
	char cmd[256], buf[8][64];
	int len;
	
	printf("> ");
	fgets(cmd, 256, stdin);
	len = split(cmd, *buf, 8, 64);
	
	if(len == 0) continue;
	
	if(len == 3 && (strcmp(buf[0], "print") == 0 || strcmp(buf[0], "p") == 0)) {
	  unsigned long val = strtol(buf[2], NULL, 0);
	  
	  printf("0x%08x\n", (strcmp(buf[1], "r") == 0) ? reg[val] : mem[val]);
	  
	} else if(len == 4 && (strcmp(buf[0], "set") == 0 || strcmp(buf[0], "s") == 0)) {
	  unsigned long val1 = strtol(buf[2], NULL, 0);
	  unsigned long val2 = strtol(buf[3], NULL, 0);
	  
	  ((strcmp(buf[1], "r") == 0) ? reg[val1] : mem[val1]) = val2;
	
	} else if(strcmp(buf[0], "continue") == 0 || strcmp(buf[0], "c") == 0) {
	  br_skip = (len > 1) ? strtol(buf[1], NULL, 0) : 0;
	  status = NULL;
	  break;
	} else if(strcmp(buf[0], "quit") == 0 || strcmp(buf[0], "q") == 0) {
	  dpc = 0;
	  break;
	} else if(strcmp(buf[0], "next") == 0 || strcmp(buf[0], "n") == 0)
	  break;
	else if(strcmp(buf[0], "help") == 0 || strcmp(buf[0], "h") == 0) {
	  puts("print|p r|m ");
	  puts("  WX^E̒lvg");
	  puts("set|s r|m  ");
	  puts("  WX^EɒlZbg");
	  puts("continue|c []");
	  puts("  sĊJ");
	  puts("  ݒ肷ƈȍ~̃u[N|Cg̐");
	  puts("quit|q");
	  puts("  s~");
	  puts("next|n");
	  puts("  ̍ss");
	}
	else
	  puts("invalid command.");
      }
    }
    /* Ȃ΁A~Ɣf */
  } while(dpc);
  
  print_report();
  
  return 0;
}

int split(char *str, char *buf, int rows, int cols) {
  int x, i, j, flag;
  
  if(str == NULL || buf == NULL || rows <= 0 || cols <= 0) return NULL;
  
  i = j = 0;
  
  for(x = 0; str[x] != NULL && str[x] != '\n'; x++) {
    switch(str[x]) {
    case ' ':
    case '\t':
      if(j) {
        buf[cols * i + j] = NULL;
        i++;
        j = 0;
      }
      break;
    default:
      if(i == rows) return NULL;
      if(j == cols - 1) return NULL;
      buf[cols * i + j] = str[x];
      j++;
    }
  }
  
  buf[cols * i + j] = NULL;
  if(j) i++;
  
  return i;
}
